"""Модуль spectacular extensions."""

from typing import Any

from drf_spectacular.extensions import OpenApiAuthenticationExtension
from drf_spectacular.openapi import AutoSchema
from drf_spectacular.plumbing import build_bearer_security_scheme_object


class EndpointAuthenticationExtension(OpenApiAuthenticationExtension):
    """Схема endpoint auth."""

    target_class = "email_auth_remote.authentication.EndpointAuthentication"
    name = "EndpointAuthentication"
    match_subclasses = True

    def get_security_definition(self, auto_schema: AutoSchema) -> Any:
        return build_bearer_security_scheme_object(
            header_name="Authorization",
            token_prefix=self.target.keyword,
        )



"""
Конфигурация приложения email_auth_remote
"""

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


class EmailAuthRemoteConfig(AppConfig):
    """
    Приложение отвечающие за email_auth_remote
    """

    default_auto_field = "django.db.models.BigAutoField"
    name = "email_auth_remote"

    def ready(self) -> None:
        """
        После загрузки приложения импортирует spectacular схему.
        Проверяет наличие AUTH_ENDPOINT_URL.
        """
        from .extensions import (  # pylint: disable=import-outside-toplevel, unused-import
            EndpointAuthenticationExtension,
        )

        if not hasattr(settings, "AUTH_ENDPOINT_URL"):
            raise ImproperlyConfigured("AUTH_ENDPOINT_URL must be set in settings.")



"""Модуль моделей для email_auth_remote."""

from functools import cached_property
from typing import Optional, List
from django.db.models.manager import EmptyManager
from django.contrib.auth import models as auth_models


class RemoteEmailUser:
    """
    Класс пользователя для email_auth_remote.EndpointAuthentication.

    Пользователя нельзя сохранить в БД.
    """

    _is_staff = False
    _is_active = True
    _is_superuser = False

    _groups = EmptyManager(auth_models.Group)
    _user_permissions = EmptyManager(auth_models.Permission)

    def __init__(self, user_id: int, email: str) -> None:
        self._id = user_id
        self._email = email

    def __str__(self) -> str:
        return f"RemoteEmailUser {self.id}"

    @cached_property
    def id(self) -> int:
        """ID пользователя."""
        return self._id

    @cached_property
    def pk(self) -> int:
        """Primary Key пользователя. В данном случае = ID."""
        return self.id

    @cached_property
    def email(self) -> str:
        """Email пользователя."""
        return self._email

    @cached_property
    def is_staff(self) -> bool:
        """
        Пользователь не staff.
        Не уверен, что доступ к админке можно получить с таким пользователем
        (пользователь для админки должен быть сохранен в БД).
        """
        return self._is_staff

    @cached_property
    def is_superuser(self) -> bool:
        """
        Пользователь не superuser.
        Не уверен, что доступ к админке можно получить с таким пользователем
        (пользователь для админки должен быть сохранен в БД).
        """
        return self._is_superuser

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RemoteEmailUser):
            return NotImplemented
        return self.id == other.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.id)

    def save(self) -> None:
        """Пользователя нельзя сохранить в БД."""
        raise NotImplementedError("Remote email users have no DB representation")

    def delete(self) -> None:
        """Пользователя нельзя сохранить в БД."""
        raise NotImplementedError("Remote email users have no DB representation")

    def set_password(self, raw_password: str) -> None:
        """Пользователя нельзя сохранить в БД."""
        raise NotImplementedError("Remote email users have no DB representation")

    def check_password(self, raw_password: str) -> None:
        """Пользователя нельзя сохранить в БД."""
        raise NotImplementedError("Remote email users have no DB representation")

    @property
    def groups(self) -> auth_models.Group:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permission are not implemented yet.")
        # return self._groups

    @property
    def user_permissions(self) -> auth_models.Permission:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permissions are not implemented yet.")
        # return self._user_permissions

    def get_group_permissions(self, obj: Optional[object] = None) -> set:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permissions are not implemented yet.")

    def get_all_permissions(self, obj: Optional[object] = None) -> set:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permissions are not implemented yet.")

    def has_perm(self, perm: str, obj: Optional[object] = None) -> bool:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permissions are not implemented yet.")

    def has_perms(self, perm_list: List[str], obj: Optional[object] = None) -> bool:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permissions are not implemented yet.")

    def has_module_perms(self, module: str) -> bool:
        """У пользователя нет Group и Permissions. WIP."""
        raise NotImplementedError("Remote auth permissions are not implemented yet.")

    @property
    def is_anonymous(self) -> bool:
        """Пользователь не анонимный."""
        return False

    @property
    def is_authenticated(self) -> bool:
        """Пользователь всегда аутентифицирован."""
        return True

    def get_username(self) -> str:
        """У пользователя email вместо имени."""
        return self._email



"""Модуль аутентификации с помощью auth endpoint."""

import logging
from typing import Optional, Tuple
from rest_framework.authentication import BaseAuthentication, get_authorization_header
from rest_framework import exceptions
from rest_framework.request import Request
from django.utils.translation import gettext_lazy as _
from django.conf import settings
import requests
from .models import RemoteEmailUser

logger = logging.getLogger(__name__)


class EndpointAuthentication(BaseAuthentication):
    """
    Аутентификация с помощью auth endpoint (settings.AUTH_ENDPOINT_URL).
    В случае успеха заменяет модель пользователя у request на RemoteEmailUser.
    """

    www_authenticate_realm = "api"
    media_type = "application/json"
    keyword = "Bearer"

    def authenticate_header(self, request: Request) -> str:
        return f'{self.keyword} realm="{self.www_authenticate_realm}"'

    def authenticate(self, request: Request) -> Optional[Tuple[RemoteEmailUser, None]]:
        auth = get_authorization_header(request)

        if not auth:
            return None

        try:
            response = requests.get(
                settings.AUTH_ENDPOINT_URL,  # type: ignore[misc]
                timeout=10,
                headers={"Authorization": auth},
            )
            response.raise_for_status()
            user_data = response.json()
            user = RemoteEmailUser(user_id=user_data["id"], email=user_data["email"])
            return user, None
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.Timeout,
        ) as error:
            logger.error(str(error))
            raise exceptions.APIException(
                detail=_("Could not connect to authentication server.")
            )
        except requests.exceptions.HTTPError as http_err:
            raise exceptions.AuthenticationFailed(detail=http_err.response.json())
        except KeyError as exc:
            raise exceptions.APIException(
                detail=_(
                    "Invalid user credentials. Maybe authentication api has changed?"
                )
            ) from exc
        except Exception as exc:
            logger.exception(str(exc))
            raise exceptions.APIException(
                detail=_("Something went wrong. Please try again.")
            )



"""
Инициализирует пакет
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("email-auth-remote")
except PackageNotFoundError:
    # package is not installed
    __version__ = None  # type: ignore[assignment]
